/*
 *
 * Copyright (c) 2016-2020 Deephaven Data Labs and Patent Pending
 */

package com.illumon.demo.sample;

import com.fishlib.configuration.Configuration;
import com.fishlib.io.log.LogLevel;
import com.fishlib.io.logger.Log4jAdapter;
import com.fishlib.io.logger.Logger;
import com.fishlib.logger.ProcessUnsafeLogCreator;
import com.fishlib.net.CommBase;
import com.fishlib.net.impl.nio.NIODriver;
import com.fishlib.util.process.ProcessEnvironment;
import com.illumon.iris.controller.PersistentQueryClient;
import com.illumon.iris.controller.PersistentQueryControllerClient;
import com.illumon.iris.controller.PersistentQueryInfo;
import com.illumon.iris.db.tables.DataColumn;
import com.illumon.iris.db.tables.Table;
import com.illumon.iris.db.tables.utils.DBTimeZone;
import com.illumon.iris.db.tables.utils.TableTools;
import io.deephaven.auth.AuthenticationClientManager;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.io.UncheckedIOException;

/**
 * Simple example for connecting to a table from a Persistent Query and returning some data.
 *
 * This example creates a Persistent Query Controller Client, and uses its methods to get a list
 * of available queries, and then to connect to one and retrieve a table from it.
 *
 * This class requires a -DAuthenticationClientManager.defaultPrivateKeyFile
 * passed as one of the JVM arguments, to provide a private key to use for authentication to Deephaven.
 * It also requires a properties file, usually named iris-common.prop, which includes the
 * details about the Persistent Query Controller used in the environment. This properties file
 * must be in the classpath when this class is executed. The name of the properties file should be
 * passed as a JVM argument with -DConfiguration.rootFile=<name_of_property_file>
 *
 * Setting up a key pair for authentication:
 * 1. Set up Deephaven Authorization keys on the Deephaven server by executing the following on
 * the server:/usr/illumon/latest/bin/generate_iris_keys <irisUserName>
 * cat pub-<irisUserName>.base64.txt >> /etc/sysconfig/illumon.d/resources/dsakeys.txt
 *
 * 2. Copy the user's private auth key ( priv-<irisUserName>.base64.txt ) to the user's home directory,
 * or somewhere else that is only accessible to the user.
 */

public class SimplePQConnection {
    private static final String PROCESS_NAME = "SimplePQConnectionDemo";
    private static Logger log;

    private static void printUsageAndExit(@NotNull final String message) {
        System.err.println(message);
        System.err.println("Usage: SimplePQConnection query_name table_name [maxRows]");
        System.exit(1);
    }

    private static void prepareEnvironment() {
        // the Configuration object reads the Deephaven property file, which has named key-value pairs
        // which control the operation of the Iris system.
        final Configuration configuration = Configuration.getInstance();

        // the log object is used by Deephaven components, and produces output in a PROCESS_NAME.log.<date> with
        // a hard link to PROCESS_NAME.log.current (e.g., DemoIntegration.log.current).  The log files are stored
        // in <workspace>/../logs
        log = new ProcessUnsafeLogCreator(PROCESS_NAME, configuration).createLogger(PROCESS_NAME, LogLevel.INFO);

        // All log4j output should be sent to the Logger that was created on the line above
        Log4jAdapter.sendLog4jToLogger(log);

        // The Deephaven system uses a wrapper around Java NIO for most network communication and job scheduling,
        // this initializes the communications subsystem with the log
        NIODriver.init(log);

        // The ProcessEnvironment provides hooks for consistent startup and shutdown initialization across
        // Deephaven processes.
        ProcessEnvironment.basicInteractiveProcessInitialization(configuration, PROCESS_NAME, log);
    }

    public static void main(@NotNull final String... args) {
        if (args.length < 2) {
            printUsageAndExit("Invalid number of arguments");
            return;
        }

        long querySerial = -1;

        // Parse parameter values from the command line
        final String pq_name = args[0];
        final String table = args[1];
        final long maxRows;
        if (args.length > 2) {
            maxRows = Long.parseLong(args[2]);
        } else {
            maxRows = 100;
        }

        final long overallStartNanos = System.nanoTime();

        // Set up logging and process management
        prepareEnvironment();

        // Create a connection to the Persistent Query Controller specified in the properties file
        final PersistentQueryControllerClient client;
        try {
            client = new PersistentQueryControllerClient(log, CommBase.getScheduler(), null, null, false);
        } catch (IOException ex) {
            throw new UncheckedIOException("Exception occurred while creating PQCClient:", ex);
        }

        // Authenticate with the Controller
        final AuthenticationClientManager.TokenFactory factory = AuthenticationClientManager.DEFAULT.getTokenFactory("PersistentQueryController");
        factory.tryActionUntil(client::authenticateToController);

        // Subscribe to the available persistent queries
        try {
            client.subscribeToAll();
        } catch (IOException ex) {
            throw new UncheckedIOException("Exception occurred while subscribing to the PQCClient:", ex);
        }

        // List all the queries and their serial numbers
        System.out.println();
        System.out.println("Available Persistent Queries:");
        for (PersistentQueryInfo info : client.values()) {
            if (info.getConfig().getName().equals(pq_name)) {
                querySerial = info.getConfig().getSerial();
            }
            System.out.println(info.getConfig().toString());
        }

        if (querySerial == -1) {
            throw new RuntimeException("Failed to find a serial number for Persistent Query: " + pq_name);
        }

        // Get a connection to the requested PQ
        final PersistentQueryInfo queryInfo = client.get(querySerial);
        final PersistentQueryClient queryClient = new PersistentQueryClient(queryInfo, log);

        // Connect to the requested table
        final long startGetNanos = System.nanoTime();
        final Table demo = queryClient.getTable(table);
        final long endGetNanos = System.nanoTime();

        /* Print out some data from the table:
        // (this will be mixed in with a lot of INFO messages if -DLoggerFactory.teeOutput=true)
        // Note that this will involve a number of round trips to the server to retrieve data
        // and may be inconsistent if the table is ticking. show() is used here simply to
        // demonstrate retrieval of data from the query table, and is not a preferred method
        // of working with remote table data.
        */
        final long startShowNanos = System.nanoTime();
        if (demo != null) {
            System.out.println();
            final String[] columns = new String[demo.getColumns().length];
            int colindex = 0;
            for (DataColumn col : demo.getColumns()) {
                columns[colindex++] = col.getName();
            }
            final long rows = demo.size() < maxRows ? demo.size() : maxRows;
            TableTools.show(demo, rows, DBTimeZone.TZ_NY, columns);
        } else {
            throw new RuntimeException("Unable to connect to table: " + table);
        }

        final long endNanos = System.nanoTime();

        // Output the timing results
        System.out.println();
        System.out.println("getTable duration " + (endGetNanos - startGetNanos) + " ns");
        System.out.println("show() table duration " + (endNanos - startShowNanos) + " ns");
        System.out.println("Overall duration " + (endNanos - overallStartNanos) + " ns");
        System.out.println();
    }
}
